using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Xml;

namespace System.Xml
{
  public static class Xexpression
  {
    private static int trace_depth = 0;

    public static int TraceDepth {
      get { return trace_depth; }
      set { trace_depth = value; }
    }

    public class Error : System.Exception
    {
      // All time errors.
      // System.OutOfMemoryException or System.InsufficientMemoryException
      public static Symbol MemoryError = "memory-error";
      // System.Exception
      public static Symbol UnknownError = "unknown-error";
      public static Symbol WrongArgument = "wrong-argument";

      // Load time errors.
      public static Symbol InvalidInteger = "invalid-integer";
      public static Symbol TermTypeInvalid = "term-type-invalid";
      public static Symbol FunctionConflict = "function-conflict";
      public static Symbol VariableTypeConfclit = "variable-type-conflict";
      public static Symbol VariableRangeConfclit = "variable-range-conflict";
      public static Symbol VariableWrongRange = "variable-wrong-range";
      public static Symbol VariableWrongValue = "variable-wrong-value";
      public static Symbol UnknownFunction = "unknown-function";
      public static Symbol MacroExpansionError = "macro-expansion-error";
      public static Symbol NoVariableName = "no-variable-anme";

      // Run time errors.
      public static Symbol ArithmeticError = "arithmetic-error";
      public static Symbol WrongType = "wrong-type";
      public static Symbol IndexOutOfRange = "index-out-of-range";
      public static Symbol ValueOutOfRange = "value-out-of-range";
      public static Symbol NoLoopToBreak = "no-loop-to-break";
      public static Symbol UncaughtThrow = "uncaught-throw";

      private readonly Symbol name;
      private Term[] stack;
      private XmlNode node;

      public Error (Symbol name, XmlNode node, string fmt, params object[] args)
	: base (String.Format (fmt, args))
	{
	  this.name = name;
	  this.node = node;
	}

      public Error (Symbol name, string fmt, params object[] args)
	: base (String.Format (fmt, args))
	{
	  this.name = name;
	}

      public Symbol Name { get { return name; } }
      public XmlNode Node {
	get { return node; }
	internal set { node = value; }
      }
      public Term[] Stack {
	get { return stack; }
	internal set { stack = value; }
      }
    }

    public struct Symbol : IEquatable<Symbol>
    {
      private static NameTable nt = new NameTable ();

      private string name;

      public Symbol (string str) { name = nt.Add (str); }

      public static implicit operator Symbol (string str)
      {
	return new Symbol (str);
      }

      public static implicit operator string (Symbol sym) { return sym.name; }

      public static bool operator== (Symbol n1, Symbol n2)
	{
	  return (object) n1.name == (object) n2.name;
	}

      public static bool operator!= (Symbol n1, Symbol n2)
	{
	  return (object) n1.name != (object) n2.name;
	}

      public static bool operator== (Symbol n1, string n2)
	{
	  return (object) n1.name == (object) n2;
	}

      public static bool operator!= (Symbol n1, string n2)
	{
	  return (object) n1.name != (object) n2;
	}

      public static bool operator== (string n1, Symbol n2)
	{
	  return (object) n1 == (object) n2.name;
	}

      public static bool operator!= (string n1, Symbol n2)
	{
	  return (object) n1 != (object) n2.name;
	}

      public bool Equals (Symbol name)
      {
	return Object.ReferenceEquals (this.name, name.name);
      }

      public override bool Equals (object obj)
      {
	return (obj is Symbol ? this == (Symbol) obj
		: obj is string ? this.name == (string) obj
		: false);
      }

      public override int GetHashCode ()
      {
	return name.GetHashCode ();
      }

      public static NameTable NameTable { get { return nt; } }

      public override string ToString () { return name; }
    }

    private static Symbol Qexpr = "expr";

    private static Symbol Qnull = "";
    private static Symbol Qfuncall = "funcall";
    private static Symbol Qinteger = "integer";
    private static Symbol Qstring = "string";
    private static Symbol Qsymbol = "symbol";
    private static Symbol Qlist = "list";

    public static Symbol Qdefun = "defun";
    public static Symbol Qdefmacro = "defmacro";
    private static Symbol Qfname = "fname";
    private static Symbol Qargs = "args";
    private static Symbol Qfixed = "fixed";
    private static Symbol Qoptional = "optional";
    private static Symbol Qrest = "rest";

    public static Symbol Qdefvar = "defvar";
    private static Symbol Qvname = "vname";
    private static Symbol Qdescription = "description";
    private static Symbol Qrange = "range";

    private static Symbol Qename = "ename";

    public class Variable
    {
      public Domain domain;
      public readonly Symbol name;
      protected Term val;

      public Variable (Domain domain, Symbol name, Term val)
      {
	this.domain = domain;
	this.name = name;
	this.val = val;
      }

      public Term Value { get { return val; } }

      public virtual Term SetValue (Term val) { this.val = val; return val; }

      public virtual Variable Clone (Domain domain)
      {
	return new Variable (domain, name, val);
      }

      public virtual void Reset () { val = Zero; }

      public override string ToString () { return name + "(" + val + ")"; }

      public abstract class Typed : Variable
      {
	protected string desc;
	private Term default_val;

	protected Typed (Domain domain, Symbol name, Term val, string desc)
	  : base (domain, name, val)
	  {
	    this.desc = desc;
	    default_val = val;
	  }

	public override void Reset () { val = default_val; }

	public string Description {
	  get { return desc; }
	  set { desc = value; }
	}

	public override Term SetValue (Term val)
	{
	  if (! CheckValue (val))
	    throw new Error (Error.VariableWrongValue,
			     "Wrong value for variable {0}: {1}",
			     name, val);
	  this.val = val;
	  return val;
	}

	public abstract bool CheckValue (Term val);
      }

      public class Int : Typed
      {
	private int[] range;

	public void IsSubrange (int[] r)
	  {
	    if (range == null)
	      return true;
	    for (int i = 0; i < r.Length; i += 2)
	      {
		int j;
		for (j = 0; j < range.Length; j += 2)
		  if (range[j] <= r[i] && range[j + 1] >= r[i + 1])
		    break;
		if (j >= range.Length)
		  return false;
	      }
	    return true;
	  }

	public Int (Domain domain, Symbol name, int n, string desc, int[] range)
	  : base (domain, name, new Term (n), desc)
	  {
	    if (range != null && range.Length % 2 == 1)
	      throw new Error (Error.VariableWrongRange,
			       "Range length for {0} not even", name);
	    this.range = range;
	    CheckValue (val);
	  }

	public override bool CheckValue (Term term)
	{
	  if (term.IsInt)
	    {
	      if (range == null)
		return true;
	      int n = term.Intval;
	      for (int i = 0; i < range.Length; i += 2)
		if (range[i] <= n && range[i + 1] >= n)
		  return true;
	    }
	  return false;
	}

	public override Variable Clone (Domain domain)
	{
	  return new Int (domain, name, val.Intval, desc, range);
	}

	public int[] Range { get { return range; } set { range = value; } }
      }

      public class Str : Typed
      {
	private string[] range;

	public bool IsSubrange (string[] r)
	  {
	    if (range == null)
	      return true;
	    for (int i = 0; i < r.Length; i++)
	      {
		int j;
		for (j = 0; j < range.Length; j++)
		  if (range[j] == r[i])
		    break;
		if (j >= range.Length)
		  return false;
	      }
	    return true;
	  }

	public Str (Domain domain, Symbol name, string str, string desc,
		    string[] range)
	  : base (domain, name, new Term (str), desc)
	  {
	    this.range = range;
	    CheckValue (val);
	  }

	public override bool CheckValue (Term val)
	{
	  if (val.IsStr)
	    {
	      if (range == null)
		return true;
	      string str = term.Strval;
	      foreach (string s in range)
		if (str == s)
		  return true;
	    }
	  return false;
	}

	public override Variable Clone (Domain domain)
	{
	  return new Str (domain, name, val.Strval, desc, range);
	}

	public string[] Range { get { return range; } set { range = value; } }
      }

      public class Sym : Typed
      {
	public Symbol[] range;

	public bool IsSubrange (Symbol[] r)
	  {
	    if (range == null)
	      return true;
	    for (int i = 0; i < r.Length; i++)
	      {
		int j;
		for (j = 0; j < range.Length; j++)
		  if (range[j] == r[i])
		    break;
		if (j >= range.Length)
		  return false;
	      }
	    return true;
	  }

	public Sym (Domain domain, Symbol name, Symbol sym, string desc,
		    Symbol[] range)
	  : base (domain, name, new Term (sym), desc)
	  {
	    this.range = range;
	    CheckValue (val);
	  }

	public override bool CheckValue (Term val)
	{
	  if (val.IsSymbol)
	    {
	      if (range == null)
		return true;
	      Symbol sym = term.Symval;
	      foreach (Symbol n in range)
		if (sym == n)
		  return true;
	    }
	  return false;
	}

	public override Variable Clone (Domain domain)
	{
	  return new Sym (domain, name, val.Symval, desc, range);
	}

	public Symbol[] Range { get { return range; } set { range = value; } }
      }
    }

    public abstract class Function
    {
      public Symbol name;
      // True iff the function can be called with vname="VNAME" attribute.
      internal readonly  bool with_var;
      // Minimum number of positional arguments.
      internal readonly int min_args;
      // If non-nengative, maximum number of arguments.  If negative,
      // maxinum number of positional arguments is (-1 - max_args),
      // and the remaining unlimit number of arguments are given as a
      // single list.
      internal readonly int max_args;

      public Function () { }

      public Function (Symbol name, bool with_var, int min_args, int max_args)
      {
	this.name = name;
	this.with_var = with_var;
	this.min_args = min_args;
	this.max_args = max_args;
      }

      public abstract Term Call (Domain domain, Variable vari, Term[] args);

      public override string ToString ()
      {
	return name;
      }

      internal class Subroutine : Function
      {
	public Func<Domain, Variable, Term[], Term> builtin;

	public Subroutine (Func<Domain, Variable, Term[], Term> builtin,
			   Symbol name, bool with_var,
			   int min_args, int max_args)
	  : base (name, with_var, min_args, max_args)
	  {
	    this.builtin = builtin;
	  }

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  if (args.Length > 0)
	    {
	      args = (Term[]) args.Clone ();
	      for (int i = 0; i < args.Length; i++)
		{
		  args[i] = args[i].Eval (domain);
		  if (domain.Thrown ())
		    return args[i];
		}
	    }
	  return builtin (domain, vari, args);
	}
      }

      internal class SpecialForm : Function
      {
	public Func<Domain, Variable, Term[], Term> builtin;

	public SpecialForm (Func<Domain, Variable, Term[], Term> builtin,
			    Symbol name, bool with_var,
			    int min_args, int max_args)
	  : base (name, with_var, min_args, max_args)
	  {
	    this.builtin = builtin;
	  }

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  return builtin (domain, vari, args);
	}
      }

      internal class Lambda : Function
      {
	private Variable[] args;
	private Term[] body;

	public Lambda (Symbol name, int min_args, int max_args, Variable[] args,
		       Term[] body)
	  : base (name, false, min_args, max_args)
	  {
	    this.args = args;
	    this.body = body;
	  }

	public void SetBody (Term[] body)
	{
	  this.body = body;
	}

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  Bindings current = domain.bindings;
	  Term result = Zero;
	  int limit = max_args >= 0 ? args.Length : args.Length - 1;

	  try {
	    for (int i = 0; i < limit; i++)
	      {
		result = args[i].Eval (domain);
		if (domain.Thrown)
		  return result;
		domain.Bind (this.args[i], result);
	      }
	    if (max_args < 0)
	      {
		List<Term> list = new List<Term> ();
		list.InsertRange (0, args[result].Listval);

		for (int i = 0; i < list.Count; i++)
		  {
		    result = list[i].Eval (domain);
		    if (domain.Thrown)
		      return result;
		    list[i] = result;
		  }
		domain.Bind (this.args[limit], new Term (list));
	      }
	    try {
	      domain.Catch (CatchTag.Return);
	      foreach (Term term in body)
		{
		  result = term.Eval (domain);
		  if (domain.Thrown ())
		    return result;
		}
	    } finally {
	      domain.Uncatch ();
	    }
	  } finally {
	    domain.UnboundTo (current);
	  }
	  return result;
	}
      }

      internal class Macro : Function
      {
	private Variable[] args;
	private Term[] body;

	public Macro (Symbol name, int min_args, int max_args, Variable[] args,
		      Term[] body)
	  : base (name, false, min_args, max_args)
	  {
	    this.args = args;
	    this.body = body;
	  }

	public void SetBody (Term[] body)
	{
	  this.body = body;
	}

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  Bindings current = domain.bindings;
	  Term result = Zero;

	  try {
	    for (int i = 0; i < args.Length; i++)
	      domain.Bind (this.args[i], args[i]);
	    try {
	      domain.Catch (CatchTag.Return);
	      foreach (Term term in body)
		{
		  result = term.Eval (domain);
		  if (domain.Thrown ())
		    break;
		}
	    } finally {
	      domain.Uncatch ();
	    }
	  } finally {
	    domain.UnboundTo (current);
	  }
	  return result;
	}
      }
    }

    internal class Bindings
    {
      private Variable vari;
      private Term old_value;
      private Bindings next;
	
      private Bindings (Variable vari)
      {
	this.vari = vari;
	old_value = vari.Value;
      }
	
      public static Bindings Bind (Bindings bindings, Variable vari, Term val)
      {
	Bindings b = new Bindings (vari);

	b.vari.Value = val;
	b.next = bindings;
	return b;
      }

      internal Bindings UnboundTo (Bindings boundary)
      {
	for (Bindings b = this; b != boundary; b = b.next)
	  b.vari.Value = b.old_value;
	return boundary;
      }

      public override string ToString ()
      {
	string str = "(bindings";
	for (Bindings b = this; b != null; b = b.next)
	  str += " " + vari.name + "=" + b.old_value;
	return str + ")";
      }
    }

    internal struct CatchTag
    {
      private Term tag;

      // Normal symbolic catch-tag.
      public CatchTag (Symbol sym) { tag = new Term (sym); }
      // Error catch-tag.
      public CatchTag (Symbol sym, string msg) { tag = new Term (sym, msg); }
      // Error catch-tag.
      public CatchTag (ErrorTerm e) { tag = new Term (e); }
      // Special catch-tag.
      private CatchTag (int i) { tag = new Term (i); }

      public Term Tag { get { return tag; } }

      public static CatchTag Return = new CatchTag (0);
      public static CatchTag Break = new CatchTag (1);
    }

    public class Domain
    {
      public Symbol name;
      public object context;
      public int depth = 0;
      private Stack<Term> call_stack = new Stack<Term> ();

      internal Dictionary<Symbol, MethodInfo> termtypes
	= new Dictionary<Symbol, MethodInfo> ();
      internal Dictionary<Symbol, Function> functions
	= new Dictionary<Symbol, Function> ();
      internal Dictionary<Symbol, Function.Macro> macros
	= new Dictionary<Symbol, Function.Macro> ();
      internal Dictionary<Symbol, Variable> variables
	= new Dictionary<Symbol, Variable> ();
      internal Bindings bindings;
      private Stack<CatchTag> catch_stack = new Stack<CatchTag> ();
      private int catch_count = 0;

      internal Domain (Symbol name) { this.name = name; }

      public Domain (Symbol name, object context)
        : this (name, basic, context) { }

      public Domain (Symbol name, Domain parent, object context) : this (name)
      {
	termtypes = new Dictionary<Symbol, MethodInfo> (parent.termtypes);
	functions = new Dictionary<Symbol, Function> (parent.functions);
	macros = new Dictionary<Symbol, Function.Macro> (parent.macros);
	variables = new Dictionary<Symbol, Variable> (parent.variables);
	this.context = context;
      }

      public int CallStackCount { get { return call_stack.Count; } }

      public void CallStackPush (Term term)
      {
	call_stack.Push (term);
      }

      public void CallStackPop ()
      {
	call_stack.Pop ();
      }

      public void CallStackPop (int count)
      {
	while (count < call_stack.Count)
	  call_stack.Pop ();
      }

      internal void Bind (Variable vari, Term value)
      {
	bindings = Bindings.Bind (bindings, vari, value);
      }

      internal void UnboundTo (Bindings boundary)
      {
	if (bindings != null)
	  bindings = bindings.UnboundTo (boundary);
      }

      internal void Catch (CatchTag tag)
      {
	catch_stack.Push (tag);
	catch_count++;
      }

      internal void Uncatch ()
      {
	catch_stack.Pop ();
	if (catch_count > catch_stack.Count)
	  catch_count--;
      }

      public bool Thrown () { return catch_count < catch_stack.Count; }

      public bool Thrown (out bool caught)
      {
	if (catch_count < catch_stack.Count)
	  {
	    caught = catch_count == catch_stack.Count - 1;
	    return true;
	  }
	caught = false;
	return false;
      }

      internal void ThrowReturn ()
      {
	foreach (CatchTag elt in catch_stack)
	  {
	    catch_count--;
	    if (elt.Tag == CatchTag.Return.Tag)
	      break;
	  }
      }

      internal bool ThrowBreak ()
      {
	if (catch_stack.Peek ().Tag != CatchTag.Break.Tag)
	  return false;
	catch_count--;
	return true;
      }

      internal void ThrowSymbol (Term tag)
      {
	foreach (CatchTag elt in catch_stack)
	  {
	    catch_count--;
	    if (elt.Tag.Matches (tag))
	      break;
	  }
      }

      internal Term ThrowError (Term tag)
      {
	tag = tag.Clone ();
	ErrorTerm e = tag.objval as ErrorTerm;
	e.CallStack = call_stack.ToArray ();

	foreach (CatchTag elt in catch_stack)
	  {
	    catch_count--;
	    if (elt.Tag.Matches (tag))
	      break;
	  }
	return tag;
      }

      public void DefType (Type type)
      {
	if (! type.IsSubclassOf (typeof (TermValue)))
	  throw new Error (Error.TermTypeInvalid,
			   "Not a subclass of TermValue: {0}", type);
	BindingFlags flags = BindingFlags.Static | BindingFlags.NonPublic;
	FieldInfo finfo = type.GetField ("name", flags);
	if (finfo == null || finfo.FieldType != typeof (Symbol))
	  throw new Error (Error.TermTypeInvalid,
			   "No \"name\" field of string type: {0}", type);
	Symbol name = (Symbol) finfo.GetValue (null);
	if (termtypes.ContainsKey (name)
	    || functions.ContainsKey (name)
	    || macros.ContainsKey (name))
	  throw new Error (Error.TermTypeInvalid,
			   "Already defined: {0}", name);
	Type[] types = new Type[] { typeof (Domain), typeof (XmlNode) };
	MethodInfo minfo = type.GetMethod ("Parser", types);
	if (minfo == null || minfo.ReturnType != typeof (TermValue))
	  throw new Error ("No \"Parser\" method of correct type: {0}", type);
	termtypes[name] = minfo;
      }

      static void check_func_head (Symbol name, bool with_var,
				   int min_args, int max_args)
      {
	if (termtypes.ContainsKey (name))
	  throw new Error (Error.FunctionConflict,
			   "Already defined as a term type: {0}", name);
	else if (macros.ContainsKey (name))
	  throw new Error (Error.FunctionConflict,
			   "Already defihed as a macro: {0}", name);
	Function func;
	if (functions.TryGetValue (name, out func)
	    && (func.with_var != with_var
		|| func.min_args != min_args
		|| func.max_args != max_args))
	  throw new Error (Error.FunctionConflict,
			   "Conflicting arguments: {0}", name);
      }

      public void DefSubr (Func<Domain, Variable, Term[], Term> builtin,
			   Symbol name, bool with_var,
			   int min_args, int max_args)
      {
	check_func_head (name, with_var, min_args, max_args);
	functions[name] = new Function.Subroutine (builtin, name, with_var,
						   min_args, max_args);
      }

      public void DefSpecial (Func<Domain, Variable, Term[], Term> builtin,
			      Symbol name, bool with_var,
			      int min_args, int max_args)
      {
	check_func_head (name, with_var, min_args, max_args);
	functions[name] = new Function.SpecialForm (builtin, name, with_var,
						    min_args, max_args);
      }

      public void Defun (Symbol name, int min_arg, int max_arg,
			 Variable[] args, Term[] body)
      {
	check_func_head (name, false, min_args, max_args);
	functions[name] = new Function.Lambda (name, min_args, max_args,
					       args, body);
      }

      public void Defun (Function func)
      {
	functions[func.name] = func;
      }

      public void Defmacro (Symbol name, int min_arg, int max_arg,
			    Variable[] args, Term[] body)
      {
	if (termtypes.ContainsKey (name))
	  throw new Error (Error.FunctionConflict,
			   "Already defined as a term type: {0}", name);
	else if (function.ContainsKey (name))
	  throw new Error (Error.FunctionConflict,
			   "Already defihed as a function: {0}", name);
	Function.Macro macro;
	if (macros.TryGetValue (name, out macro)
	    && (macro.min_args != min_args
		|| macro.max_args != max_args))
	  throw new Error (Error.FunctionConflict,
			   "Conflicting arguments: {0}", name);
	macros[name] = new Function.Macro (name, min_arg, max_arg, args, body);
      }

      public void DefAlias (Symbol alias, Symbol fname)
      {
	Function func;
	if (termtypes.TryGetValue (alias, out func))
	  throw new Error (Error.FunctionConflict,
			   "Alread defined as a term type: {0}", alias);
	if (functions.TryGetValue (alias, out func))
	  throw new Error (Error.FunctionConflict,
			   "Alread defined as a function: {0}", alias);
	if (macros.TryGetValue (alias, out func))
	  throw new Error (Error.FunctionConflict,
			   "Alread defined as a macro: {0}", alias);
	if (! functions.TryGetValue (fname, out func))
	  throw new Error (Error.UnknownFunction, fname);
	functions[alias] = func;
      }

      public Variable Defvar (Variable vari)
      {
	vari = vari.Clone (this);
	variables[vari.name] = vari;
	return vari;
      }

      internal Variable Defvar (Symbol name)
      {
	Variable vari = new Variable (this, name, Zero);
	variables[name] = vari;
	return vari;
      }

      public Variable DefvarInt (Symbol name, int n, string desc, int[] range)
      {
	Variable vari;

	if (variables.TryGetValue (name, out vari))
	  {
	    Variable.Int intvari = vari as Variable.Int;
	    if (intvari == null)
	      throw new Error (Error.VariableTypeConfclit,
			       "Not an integer variable: {0}", name);
	    if (range != null)
	      {
		if (! intvari.IsSubrange (range))
		  throw new Error (Error.VariableRangeConfclit,
				   "Conflicting variable range: {0}", name);
		intvari.Range = range;;
	      }
	    if (desc != null)
	      intvari.Description = desc;
	  }
	else
	  {
	    vari = new Variable.Int (this, name, n, desc, range);
	    variables[name] = vari;
	  }
	return vari;
      }

      public Variable DefvarStr (Symbol name, string str, string desc,
				 string[] range)
      {
	Variable vari;

	if (variables.TryGetValue (name, out vari))
	  {
	    Variable.Str strvari = vari as Variable.Str;
	    if (strvari == null)
	      throw new Error (Error.VariableTypeConflict,
			       "Not a string variable: {0}", name);
	    if (range != null)
	      {
		if (! strvari.IsSubrange (range))
		  throw new Error (Error.VariableRangeConflict,
				   "Conflicting variable range: {0}", name);
		strvari.Range = range;
	      }
	    if (desc != null)
	      strvari.Description = desc;
	  }
	else
	  {
	    vari = new Variable.Str (this, name, str, desc, range);
	    variables[name] = vari;
	  }
	return vari;
      }

      public Variable DefvarSym (Symbol name, Symbol sym, string desc,
				 Symbol[] range)
      {
	Variable vari;

	if (variables.TryGetValue (name, out vari))
	  {
	    Variable.Sym symvari = vari as Variable.Sym;
	    if (symvari == null)
	      throw new Error (Error.VariableTypeConflict,
			       "Not a symbol variable: {0}", name);
	    if (range != null)
	      {
		if (! symvari.IsSubrange (range))
		  throw new Error (Error.VariableRangeConflict,
				   "Conflicting variable range: {0}", name);
		symvari.Range = range;
	      }
	    if (desc != null)
	      symvari.Description = desc;
	  }
	else
	  {
	    vari = new Variable.Sym (this, name, sym, desc, range);
	    variables[name] = vari;
	  }
	return vari;
      }

      internal Function GetFunc (Symbol name)
      {
	Function func;

	if (! functions.TryGetValue (name, out func))
	  throw new Error (Error.UnknownFunction,
			   "Unknown function: {0}", name);
	return func;
      }

      public bool CopyFunc (Domain domain, Symbol name)
      {
	Function func = GetFunc (name);

	domain.functions[name] = func;
	return true;
      }

      public void CopyFunc (Domain domain)
      {
	foreach (KeyValuePair<Symbol, Function> kv in functions)
	  domain.functions[kv.Key] = kv.Value;
      }

      public Variable GetVar (Symbol name, bool create)
      {
	Variable vari;

	if (! variables.TryGetValue (name, out vari))
	  {
	    if (! create)
	      return null;
	    variables[name] = vari = new Variable (this, name, Zero);
	  }
	return vari;
      }

      public override string ToString ()
      {
	string str = "<(functions";
	foreach (KeyValuePair<Symbol, Function> kv in functions)
	  str += " " + kv.Key;
	str += ") (variabls";
	foreach (KeyValuePair<Symbol, Variable> kv in variables)
	  str += " " + kv.Value;
	str += ")";
	if (bindings != null)
	  str += " " + bindings;
	if (context != null)
	  str += " (" + context + ")";
	str += ">";
	return str;
      }

      internal void TraceWrite (string fmt, params string[] arg)
      {
	if (trace_depth > depth)
	  {
	    for (int i = 0; i < depth; i++)
	      Console.Write ("  ");
	    Console.WriteLine (fmt, arg);
	  }
      }

      public object SaveValues ()
      {
	Dictionary<Variable,Term> values = new Dictionary<Variable,Term> ();

	foreach (KeyValuePair<Symbol,Variable> kv in variables)
	  values[kv.Value] = kv.Value.Value.Clone ();
	return values;
      }

      public void RestoreValues (object values)
      {
	foreach (KeyValuePair<Variable,Term> kv
		 in (Dictionary<Variable,Term>) values)
	  kv.Key.Value = kv.Value;
      }
    }

    private static Domain basic = new Domain ("basic");

    static Xexpression ()
    {
      basic.DefType (typeof (Varref));
      basic.DefType (typeof (Funcall));
      basic.DefType (typeof (ErrorTerm));

      basic.DefSubr (Fset, "set", true, 1, 1);
      basic.DefAlias ("=", "set");
      basic.DefSubr (Fnot, "not", false, 1, 1);
      basic.DefAlias ("!", "not");
      basic.DefSubr (Fadd, "add", true, 1, -1);
      basic.DefAlias ("+", "add");
      basic.DefSubr (Fmul, "mul", true, 1, -1);
      basic.DefAlias ("*", "mul");
      basic.DefSubr (Fsub, "sub", true, 1, -1);
      basic.DefAlias ("-", "sub");
      basic.DefSubr (Fdiv, "div", true, 1, -1);
      basic.DefAlias ("/", "div");
      basic.DefSubr (Fmod, "mod", true, 1, 2);
      basic.DefAlias ("%", "mod");
      basic.DefSubr (Flogior, "logior", true, 1, -1);
      basic.DefAlias ("|", "logior");
      basic.DefSubr (Flogand, "logand", true, 1, -1);
      basic.DefAlias ("&", "logand");
      basic.DefSubr (Flsh, "lsh", true, 1, 2);
      basic.DefAlias ("<<", "lsh");
      basic.DefSubr (Frsh, "rsh", true, 1, 2);
      basic.DefAlias (">>", "rsh");
      basic.DefSubr (Feq, "eq", false, 2, -1);
      basic.DefAlias ("==", "eq");
      basic.DefSubr (Fnoteq, "noteq", false, 2, 2);
      basic.DefAlias ("!=", "noteq");
      basic.DefSubr (Flt, "lt", false, 2, -1);
      basic.DefAlias ("<", "lt");
      basic.DefSubr (Fle, "le", false, 2, -1);
      basic.DefAlias ("<=", "le");
      basic.DefSubr (Fgt, "gt", false, 2, -1);
      basic.DefAlias (">", "gt");
      basic.DefSubr (Fge, "ge", false, 2, -1);
      basic.DefAlias (">=", "ge");

      basic.DefSubr (Fappend, "append", true, 0, -1);
      basic.DefSubr (Fconcat, "concat", true, 0, -1);
      basic.DefSubr (Fnth, "nth", false, 2, 2);
      basic.DefSubr (Fcopy, "copy", false, 1, 1);
      basic.DefSubr (Fins, "ins", true, 2, 2);
      basic.DefSubr (Fdel, "del", true, 2, 2);
      basic.DefSubr (Feval, "eval", false, 1, 1);
      basic.DefSubr (Fbreak, "break", false, 0, 1);
      basic.DefSubr (Freturn, "return", false, 0, 1);
      basic.DefSubr (Fthrow, "throw", false, 1, 2);

      basic.DefSpecial (Fand, "and", false, 1, -1);
      basic.DefAlias ("&&", "and");
      basic.DefSpecial (For, "or", false, 1, -1);
      basic.DefAlias ("||", "or");
      basic.DefSpecial (Fprogn, "progn", false, 0, -1);
      basic.DefAlias ("expr", "progn");
      basic.DefSpecial (Fif, "if", false, 2, 3);
      basic.DefSpecial (Fwhen, "when", false, 1, -1);
      basic.DefSpecial (Floop, "loop", false, 1, -1);
      basic.DefSpecial (Fwhile, "while", false, 1, -1);
      basic.DefSpecial (Fcond, "cond", false, 1, -1);
      basic.DefSpecial (Fforeach, "foreach", true, 2, -1);
      basic.DefSpecial (Fquote, "quote", false, 1, 1);
      basic.DefSpecial (Ftype, "type", false, 1, 1);
      basic.DefSpecial (Fcatch, "catch", true, 2, -1);
    }

    private static Term Fset (Domain domain, Variable vari, Term[] args)
    {
      if (vari == null)
	throw new Error (Error.NoVariableName, "No variable name to set"));
      vari.SetValue (args[0]);
      return args[0];
    }

    private static Term Fnot (Domain domain, Variable vari, Term[] args)
    {
      return args[0].IsTrue ? Zero : One;
    }

    static Term maybe_set_intvar (Variable vari, int n)
    {
      Term term = new Term (n);
      if (vari != null)
	vari.SetValue (term);
      return term;
    }


    private static Term Fadd (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 0 : vari.Value.Intval;

      foreach (Term arg in args)
	n += arg.Intval;
      return maybe_set_intvar (vari, n);
    }

    private static Term Fmul (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 1 : vari.Value.Intval;
      foreach (Term arg in args)
	n *= arg.Intval;
      return maybe_set_intvar (vari, n);
    }

    private static Term Fsub (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.Value.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n -= args[i++].Intval;
      return maybe_set_intvar (vari, n);
    }

    private static Term Fdiv (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.Value.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n /= args[i++].Intval;
      return maybe_set_intvar (vari, n);
    }

    private static Term Fmod (Domain domain, Variable vari, Term[] args)
    {
      return maybe_set_intvar (vari, args[0].Intval % args[1].Intval);
    }

    private static Term Flogior (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 0 : vari.Value.Intval;
      foreach (Term arg in args)
	n |= arg.Intval;
      return maybe_set_intvar (vari, n);
    }

    private static Term Flogand (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.Value.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n &= args[i++].Intval;
      return maybe_set_intvar (vari, n);
    }

    private static Term Flsh (Domain domain, Variable vari, Term[] args)
    {
      return maybe_set_intvar (vari, args[0].Intval << args[1].Intval);
    }

    private static Term Frsh (Domain domain, Variable vari, Term[] args)
    {
      return maybe_set_intvar (vari, args[0].Intval >> args[1].Intval);
    }

    private static Term Feq (Domain domain, Variable vari, Term[] args)
    {
      Term o = args[0];

      if (o.objval == null)
	{
	  for (int i = 1; i < args.Length; i++)
	    if (args[i].objval != null || args[i].intval != o.intval)
	      return Zero;
	}
      else
	{
	  for (int i = 1; i < args.Length; i++)
	    if (o.objval.Equals (args[i].objval))
	      return Zero;
	}
      return One;
    }

    private static Term Fnoteq (Domain domain, Variable vari, Term[] args)
    {
      return Feq (domain, vari, args);
    }

    private static Term Flt (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;

      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n >= n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fle (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n > n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fgt (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n <= n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fge (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n < n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fappend (Domain domain, Variable vari, Term[] args)
    {
      List<Term> list;

      if (vari == null)
	list = new List<Term> ();
      else
	list = vari.Value.Listval;

      foreach (Term arg in args)
	{
	  if (arg.IsList)
	    list.AddRange ((List<Term>) arg.objval);
	  else
	    list.Add (arg);
	}
      if (vari == null)
	{
	  Term result;
	  result.intval = 0;
	  result.objval = list;
	  return result;
	}
      return vari.Value;
    }

    private static Term Fconcat (Domain domain, Variable vari, Term[] args)
    {
      string str;

      if (vari == null)
	str = "";
      else
	str = vari.Value.Strval;

      foreach (Term arg in args)
	{
	  if (arg.IsStr)
	    str += (string) arg.objval;
	  else if (arg.IsList)
	    foreach (Term term in (List<Term>) arg.objval)
	      str += (char) term.Intval;
	  else
	    str += (char) arg.Intval;
	}

      Term strterm = new Term (str);
      if (vari != null)
	vari.SetValue (strterm);
      return strterm;
    }

    private static Term Fnth (Domain domain, Variable vari, Term[] args)
    {
      Term result;

      if (args[1].IsStr)
	{
	  result.intval = ((string) args[1].objval)[args[0].Intval];
	  result.objval = null;
	}
      else if (args[1].IsList)
	{
	  result = ((List<Term>) args[1].objval)[args[0].Intval];
	}
      else
	throw new Error (Error.WrongType, 
			 "Not a string nor a list: {0}", args[1]);
      return result;
    }

    private static Term Fcopy (Domain domain, Variable vari, Term[] args)
    {
      Term result;

      result.intval = 0;
      result.objval = new List<Term> (args[0].Listval);
      return result;
    }

    private static Term Fins (Domain domain, Variable vari, Term[] args)
    {
      if (vari == null)
	throw new Error (Error.NoVariableName, "No variable name to set");
      Term term = vari.Value;
      if (term.IsStr)
	term.objval = term.Strval.Insert (args[0].Intval, args[1].Strval);
      else if (vari.Value.IsList)
	vari.Value.Listval.InsertRange (args[0].Intval, args[1].Listval);
      else
	throw new Error (Error.WrongType, "Not a string nor a list: {0}",
			 vari.Value);
      return term;
    }

    private static Term Fdel (Domain domain, Variable vari, Term[] args)
    {
      if (vari == null)
	throw new Error (Error.NoVariableName, "No variable name to set");
      Term term = vari.Value;
      if (term.IsStr)
	term.objval = term.Strval.Remove (args[0].Intval,
					  args[1].Intval - args[0].Intval);
      else if (vari.Value.IsList)
	vari.Value.Listval.RemoveRange (args[0].Intval,
				      args[1].Intval - args[0].Intval);
      else
	throw new Error (Error.WrongType, "Not a string nor a list: {0}",
			 vari.Value);
      return term;
    }

    private static Term Fand (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	if (! arg.Eval (domain).IsTrue)
	  return Zero;
      return One;
    }

    private static Term For (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	if (arg.Eval (domain).IsTrue)
	  return One;
      return Zero;
    }

    private static Term Feval (Domain domain, Variable vari, Term[] args)
    {
      return (args[0].Eval (domain));
    }

    private static Term Fprogn (Domain domain, Variable vari, Term[] args)
    {
      Term result = One;

      foreach (Term arg in args)
	result = arg.Eval (domain);
      return result;
    }

    private static Term Fif (Domain domain, Variable vari, Term[] args)
    {
      if (args[0].Eval (domain).IsTrue)
	return args[1].Eval (domain);
      if (args.Length == 2)
	return Zero;
      return args[2].Eval (domain);
    }

    private static Term Fwhen (Domain domain, Variable vari, Term[] args)
    {
      if (! args[0].Eval (domain).IsTrue)
	return Zero;
      Term result = One;
      for (int i = 1; i < args.Length; i++)
	result = args[i].Eval (domain);
      return result;
    }

    private static Term Freturn (Domain domain, Variable vari, Term[] args)
    {
      domain.ThrowReturn ();
      return args.Length == 0 ? Zero : args[0];
    }	  

    private static Term Fbreak (Domain domain, Variable vari, Term[] args)
    {
      if (! domain.ThrowBreak ())
	return domain.ThrowError (new Term (Error.NoLoopToBreak, null));
      return args.Length == 0 ? Zero : args[0];
    }	  

    private static Term Floop (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (CatchTag.Break);
	while (! domain.Thrown ())
	  foreach (Term arg in args)
	    {
	      result = arg.Eval (domain);
	      if (domain.Thrown ())
		return result;
	    }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fwhile (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (CatchTag.Break);
	while (! domain.Thrown () && args[0].Eval (domain).IsTrue)
	  for (int i = 1; i < args.Length; i++)
	    {
	      result = args[i].Eval (domain);
	      if (domain.Thrown ())
		return result;
	    }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fcond (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	{
	  List<Term> list = arg.Listval;
	  Term result = list[0].Eval (domain);

	  if (result.IsTrue)
	    {
	      for (int i = 1; i < list.Count; i++)
		{
		  domain.depth++;
		  result = list[i].Eval (domain);
		  domain.depth--;
		  if (domain.Thrown ())
		    return result;
		}		  
	      return result;
	    }
	}
      return Zero;
    }

    private static Term Fforeach (Domain domain, Variable vari, Term[] args)
    {
      Term result = args[0].Eval (domain);
      if (domain.Thrown ())
	return result;
      List<Term> list = result.Listval;
      Bindings current = domain.bindings;

      try {
	domain.Catch (CatchTag.Break);
	foreach (Term term in list)
	  {
	    domain.Bind (vari, term);
	    try {
	      for (int i = 1; i < args.Length; i++)
		{
		  result = args[i].Eval (domain);
		  if (domain.Thrown ())
		    return result;
		}
	    } finally {
	      domain.UnboundTo (current);
	    }
	  }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fquote (Domain domain, Variable vari, Term[] args)
    {
      return args[0];
    }

    private static Term Ftype (Domain domain, Variable vari, Term[] args)
    {
      if (args[0].IsInt)
	return TermInt;
      if (args[0].IsStr)
	return TermStr;
      if (args[0].IsSymbol)
	return TermSymbol;
      if (args[0].IsList)
	return TermList;
      return TermTerm;
    }

    private static Term eval_terms (Domain domain, Term[] terms, int idx,
				    out bool caught)
    {
      Term result= Zero;

      caught = false;
      for (int i = idx; i < terms.Length; i++)
	{
	  result = terms[i].Eval (domain);
	  if (domain.Thrown (out caught))
	    break;
	}
      return result;
    }

    public static Term Fcatch (Domain domain, Variable vari, Term[] args)
    {
      Term result;
      bool caught = false;

      if (args.Length < 2)
	throw new Error (Error.WrongArgument,
			 "Too few arguments: {0}", args.Length);
      ErrorTerm err = args[0].objval as ErrorTerm;
      if (err != null)
	{
	  try {
	    result = eval_terms (domain, args, 1, out caught);
	  } catch (Error e) {
	    if (! args[0].Matches (e.ErrorTerm))
	      throw e;
	    if (vari != null)
	      vari.SetValue (e.ErrorTerm);
	    return One;
	  } catch (System.Exception e) {
	    Term term = new Term (new ErrorTerm (e));
	    if (! args[0].Matches (term))
	      throw e;
	    if (vari != null)
	      vari.SetValue (term);
	    return One;
	  }
	}
      else if (args[0].IsSymbol)
	{
	  domain.Catch (new CatchTag (args[0].Symval));
	  result = eval_terms (domain, args, 1, out caught);
	  domain.Uncatch ();
	}
      else
	throw new Error (Error.WrongArgument,
			 "Not a symbol nor an error: {0}", args[0]);

      if (vari != null)
	vari.SetValue (result);
      return Zero;
    }

    public static Term Fthrow (Domain domain, Variable vari, Term[] args)
    {
      if (args[0].IsSymbol)
	{
	  domain.ThrowSymbol (args[0]);
	  return (args[args.Length - 1]);
	}
      if (args[0].IsError)
	return domain.ThrowError (args[0]);
      throw new Error (Error.WrongArgument,
		       "Not a symbol nor an error: {0}", args[0]);
    }

    public abstract class TermValue
    {
      public virtual Term Eval (Domain domain) { return this; }
      public virtual TermValue Clone () { return this; }
      public virtual bool IsTrue { get { return true; } }
      public virtual bool Matches (TermValue other) { return Equals (other); }
      public override abstract bool Equals (object obj);
      public override abstract int GetHashCode ();
      public abstract string ToString (bool detail);
    }

    private class Varref : TermValue
    {
      private static Symbol name = "varref";
      public static Symbol Name { get { return name; } }

      private Symbol vname;
      private Variable vari;

      public Varref (Symbol vname) { this.vname = vname; }

      public override Term Eval (Domain domain)
      {
	if (vari == null || vari.domain != domain)
	  vari = domain.GetVar (vname, true);
	return vari.Value;
      }

      public static TermValue Parser (Domain domain, XmlNode node)
      {
	return new Varref ((Symbol) node.Attributes[Qvname].Value);
      }

      public override bool Equals (object obj)
      {
	Varref varref = obj as Varref;
	return (varref != null && varref.vari.name == vari.name);
      }

      public override int GetHashCode ()
      {
	return vari.name.GetHashCode ();
      }	

      public override string ToString () { return ToString (true); }

      public override string ToString (bool detail)
      {
	return "<varref vname=\"" + vname + "\"/>";
      }
    }

    private class Funcall : TermValue
    {
      private static Symbol name = "funcall";
      public static Symbol Name { get { return name; } }

      internal Function func;
      internal Variable vari;
      internal Term[] args;

      public Funcall (Function func, Variable vari, Term[] args)
	{
	  int nargs = args.Length;

	  if (nargs < func.min_args
	      || (func.max_args >= 0 && nargs > func.max_args))
	    throw new Error (Error.WrongArgument,
			     "Wrong number of arguments to {0}: {1}",
			     func.name, nargs);
	  if (! func.with_var && vari != null)
	    throw new Error (Error.WrongArgument,
			     "Wrong vname attribute to {0}: {1}",
			     func.name, vari.name);
	  this.func = func;
	  this.vari = vari;
	  this.args = args;
	}

      public static TermValue Parser (Domain domain, XmlNode node)
      {
	Symbol fname = node.Name;
	XmlAttribute attr;

	if (fname == Qfuncall)
	  fname = node.Attributes[Qfname].Value;
	Function func = domain.GetFunc (fname);
	Variable vari;
	attr = node.Attributes[Qvname];
	vari = attr == null ? null : domain.GetVar (attr.Value, true);
	XmlNodeList nlist = node.ChildNodes;
	int nargs = nlist.Count;
	Term[] args = new Term[nargs];
	for (int i = 0; i < nargs; i++)
	  args[i] = new Term (domain, nlist[i]);
	return new Funcall (func, vari, args);
      }

      public override Term Eval (Domain domain)
      {
	domain.TraceWrite (ToString ());
	domain.depth++;
	Term result = func.Call (domain, vari, args);
	domain.depth--;
	domain.TraceWrite ("=> {0}", result.ToString ());
	return result;
      }

      public override TermValue Clone ()
      {
	return new Funcall (func, vari, args);
      }

      public override bool Equals (object obj)
      {
	Funcall funcall = obj as Funcall;
	return (obj != null
		&& funcall.func == func
		&& funcall.vari == vari
		&& funcall.args.Length == args.Length);
      }

      public override int GetHashCode ()
      {
	return func.GetHashCode () ^ args.Length;
      }

      public override string ToString () { return ToString (true); }

      public override string ToString (bool detail)
      {
	string str = "<" + func.name;
	if (vari != null)
	  str += " vname=\"" + vari.name + "\"";
	if (args.Length == 0 || ! detail)
	  return str + "/>";
	str += ">";
	if (func is Function.SpecialForm)
	  {
	    for (int i = 0; i < args.Length; i++)
	      str += ".";
	  }
	else
	  foreach (Term e in args)
	    str += e;
	return (str + "</" + func.name + ">");
      }
    }

    internal class ErrorTerm : TermValue
    {
      private static Symbol name = "error";
      public static Symbol Name { get { return name; } }
      
      private readonly Symbol ename;
      private string message;
      private Term[] stack;

      public ErrorTerm (Symbol ename, string message)
	{
	  this.ename = ename;
	  this.message = message;
	}

      internal ErrorTerm (Exception e)
	{
	  if (e is ArithmeticException)
	    ename = Error.ArithmeticException;
	  else if (e is IndexOutOfRangeException)
	    ename = Error.IndexOutOfRange;
	  else if (e is InsufficientMemoryException
		   || e is OutOfMemoryException)
	    ename = Error.MemoryError;
	  else
	    ename = Error.UnknownError;
	  message = e.Message;
	}

      public static TermValue Parser (Domain domain, XmlNode node)
      {
	XmlAttribute attr = node.Attributes[Qename];
	Symbol ename = attr == null ? Qnull : (Symbol) attr.Value;
	string message = node.InnerText;
	return new ErrorTerm (ename, message);
      }

      public Term[] CallStack {
	get { return stack; }
	internal set { stack = value; }
      }

      public override TermValue Clone ()
      {
	return new ErrorTerm (ename, message);
      }

      public override bool Equals (object obj)
      {
	ErrorTerm e = obj as ErrorTerm;
	return (e != null && e.ename = ename && e.message != message
		&& (e.stack == null ? stack == null
		    : (stack != null && e.stack.Length == stack.Length)))
      }

      public override int GetHashCode ()
      {
	return ename.GetHashCode ();
      }

      public override bool Matches (TermValue term)
      {
	ErrorTerm e = term as ErrorTerm;
	return (e != null && (ename == Qnull || e.ename == ename));
      }

      public override string ToString () { return ToString (true); }

      public override string ToString (bool detail)
	{
	  string str = "<error type=\"" + ename + "\">";
	  if (message != null)
	    str += message;
	  if (stack != null)
	    for (int i = 0; i < stack.Length; i++)
	      str += stack[i].ToString (false);
	  return str + "</error>";
	}
    }

    public struct Term
    {
      public int intval;
      public object objval;

      // <integer>...</integer>
      public Term (int i) { intval = i; objval = null; }
      // <symbol>...</symbol>
      public Term (Symbol name) { intval = 0; objval = name; }
      // <string>...</string>
      public Term (string str) { intval = 0; objval = str; }
      // <list>...</list>
      public Term (List<Term> list) { intval = 0; objval = list; }
      // <error ename="ERROR-NAME">ERROR-MESSASGE</error>
      public Term (Symbol name, string message)
	{
	  intval = 0;
	  objval = new ErrorTerm (name, message);
	}

      public Term (TermValue obj) { intval = 0; objval = obj; }

      public Term (Domain domain, XmlNode node)
	{
	  Symbol name = node.Name;

	  if (name == Qinteger)
	    {
	      intval = parse_integer (node);
	      objval = null;
	    }
	  else
	    {
	      intval = 0;
	      if (name == Qsymbol)
		objval = (Symbol) node.InnerText;
	      else if (name == Qstring)
		objval = node.InnerText.Clone ();
	      else if (name == Qlist)
		{
		  List<Term> list = new List<Term> ();
		  for (node = node.FirstChild; node != null;
		       node = node.NextSibling)
		    list.Add (new Term (domain, node));
		  objval = list;
		}
	      else
		{
		  MethodInfo minfo;

		  if (domain.termtypes.TryGetValue (name, out minfo))
		    objval = (TermValue) minfo.Invoke (null, new Object[] { domain, node});
		  else
		    {
		      Funcall funcall = (Funcall) Funcall.Parser (domain, node);
		      if (funcall.func is Function.Macro)
			{
			  Term result = funcall.Eval (domain);
			  intval = result.intval;
			  objval = result.objval;
			}
		      else
			objval = funcall;
		    }
		}
	    }
	}

      // <varref vname="VNAME"/>
      public Term (Domain domain, Symbol vname)
	{
	  intval = 0;
	  objval = new Varref (vname);
	}

      // <funcall fname="FNAME">...</funcall>
      public Term (Domain domain, Symbol fname, Term[] args)
	: this (domain, fname, Qnull, args) { }

      // <funcall fname="FNAME" vname="VNAME">...</funcall>
      public Term (Domain domain, Symbol fname, Symbol vname, Term[] args)
	{
	  intval = 0;

	  Function func = domain.GetFunc (fname);
	  Variable vari = vname == Qnull ? null : domain.GetVar (vname, true);
	  Funcall funcall = new Funcall (func, vari, args);
	  if (func is Function.Macro)
	    {
	      Term result = funcall.Eval (domain);
	      intval = result.intval;
	      objval = result.objval;
	    }
	  else
	    objval = funcall;
	}

      public object Objval {
	get {
	  if (objval == null)
	    throw new Exception (Qwrong_type,
				 "term is an integer: {0}", this);
	  return objval;
	}
      }

      public int Intval {
	get {
	  if (objval != null)
	    throw new Exception (Qwrong_type, "{0} is not an integer", this);
	  return intval;
	}
      }

      public string Strval {
	get {
	  if (! IsStr)
	    throw new Exception (Qwrong_type, "{0} is not a string", this);
	  return (string) objval;
	}
      }

      public string Symval {
	get {
	  if (! IsSymbol)
	    throw new Exception (Qwrong_type, "{0} is not a symbol", this);
	  return (Symbol) objval;
	}
      }

      public List<Term> Listval {
	get {
	  if (! IsList)
	    throw new Exception (Qwrong_type, "{0} is not a list");
	  return (List<Term>) objval;
	}
      }

      public bool IsTrue {
	get {
	  return (objval == null
		  ? (intval != 0)
		  : objval is List<Term>
		  ? (((List<Term>) objval).Count != 0)
		  : objval is TermValue
		  ? ((TermValue) objval).IsTrue
		  : true);
	}
      }

      public bool IsInt { get { return (objval == null); } }
      public bool IsStr { get { return (objval is string); } }
      public bool IsSymbol { get { return (objval is Symbol); } }
      public bool IsList { get { return (objval is List<Term>); } }
      public bool IsError { get { return (objval is ErrorTerm); } }

      internal Term Eval (Domain domain)
      {
	if (objval == null || objval is Symbol || objval is string
	    || objval is List<Term>)
	  return this;
	domain.CallStackPush (this);
	Term result;
	try {
	  result = ((TermValue) objval).Eval (domain);
	} finally {
	  domain.CallStackPop ();
	}
	return result;
      }

      public Term Clone ()
      {
	if (objval == null || objval is Symbol || objval is string)
	  return this;
	if (objval is List<Term>)
	  {
	    List<Term> list = new List<Term> ();
	    list.InsertRange (0, ((List<Term>) objval));
	    return new Term (list);
	  }
	return new Term (((TermValue) objval).Clone ());
      }

      public static bool operator== (Term t1, Term t2)
	{
	  return t1.Equals (t2);
	}

      public static bool operator!= (Term t1, Term t2) { return ! (t1 == t2); }

      public override bool Equals (object obj)
      {
	if (! (obj is Term))
	  return false;
	Term term = (Term) obj;
	if (IsInt)
	  return term.IsInt && term.Intval == Intval;
	if (IsStr)
	  return term.IsStr && term.Strval == Strval;
	if (IsSymbol)
	  return term.IsSymbol && term.Symval == Symval;
	if (IsList)
	  {
	    if (! term.IsList)
	      return false;
	    List<Term> l1 = objval as List<Term>;
	    List<Term> l2 = term.objval as List<Term>;

	    if (l1.Count != l2.Count)
	      return false;
	    for (int i = 0; i < l1.Count; i++)
	      if (! l1[i].Equals (l2[i]))
		return false;
	    return true;
	  }
	return ((TermValue) objval).Equals (obj);
      }

      public override int GetHashCode () { return intval; }

      public bool Matches (Term term)
      {
	if (IsInt)
	  return term.IsInt && term.Intval == Intval;
	if (IsStr)
	  return term.IsStr && term.Strval == Strval;
	if (IsSymbol)
	  return term.IsSymbol && term.Symval == Symval;
	if (IsList)
	  {
	    if (! term.IsList)
	      return false;
	    List<Term> l1 = objval as List<Term>;
	    List<Term> l2 = term.objval as List<Term>;

	    if (l1.Count > l2.Count)
	      return false;
	    for (int i = 0; i < l1.Count; i++)
	      if (! l1[i].Equals (l2[i]))
		return false;
	    return true;
	  }
	return ((TermValue) objval).Matches ((TermValue) term.objval);
      }

      public override string ToString () { return ToString (true); }

      public string ToString (bool detail)
      {
	string str;

	if (objval == null)
	  str = "<integer>" + intval + "</integer>";
	else if (objval is Symbol)
	  str = "<symbol>" + objval + "</symbol>";
	else if (objval is string)
	  str = "<string>" + objval + "</string>";
	else if (objval is List<Term>)
	  {
	    if (detail)
	      {
		str = "<list>";
		foreach (Term e in (List<Term>) objval)
		  str += e;
		str += "</list>";
	      }
	    else
	      str = "<list/>";
	  }
	else if (objval is Term)
	  str = "<quote>" + objval + "</quote>";
	else
	  str = ((TermValue) objval).ToString (detail);
	return str;
      }
    }

    private static Term Zero = new Term (0);
    private static Term One = new Term (1);
    private static Term TermInt = new Term (Qinteger);
    private static Term TermStr = new Term (Qstring);
    private static Term TermSymbol = new Term (Qsymbol);
    private static Term TermList = new Term (Qlist);
    private static Term TermTerm = new Term ((Symbol) "term");

    private static int parse_integer (XmlNode node)
    {
      string str = node.InnerText;
      int len = str.Length;
      int i, c;

      if (len == 1 && str[0] == '?')
	return str[1];
      if (len > 2
	  && str[1] == 'x' && ((c = str[0]) == '0' || c == '#'))
	{
	  if (! int.TryParse (str.Substring (2),
			      Globalization.NumberStyles.HexNumber, null,
			      out i))
	    throw new Error (Error.InvalidInteger, node,
			     "Not a integer string: {0}", str);
	  return i;
	}
      if (! int.TryParse (str, out i))
	throw new Error (Error.InvalidInteger, node,
			 "Not a integer string: {0}", str);
      return i;
    }

    private static Symbol parse_defvar (Domain domain, XmlNode node)
    {
      XmlAttribute name = node.Attributes[Qvname];
      if (name == null)
	throw new Error (Error.NoVariableName, node,
			 "No variable name to define");
      Symbol vname = name.Value;
      String desc;
      Variable vari;

      node = node.FirstChild;
      if (node != null && node.Name == Qdescription)
	{
	  desc = node.InnerText;
	  node = node.NextSibling;
	}
      else
	desc = null;
      if (node != null)
	{
	  Symbol type = node.Name;
	  XmlNodeList range_list = null;
	  int nranges = 0;
	  XmlNode n = node.NextSibling;

	  if (n != null)
	    {
	      range_list = n.ChildNodes;
	      nranges = range_list.Count;
	    }

	  if (type == Qinteger)
	    {
	      int val = parse_integer (node);
	      int[] range = null;
	      if (range_list != null)
		{
		  range = new int[nranges * 2];
		  for (int i = 0; i < nranges; i++)
		    {
		      XmlNode nd = range_list[i];
		      if (nd.Name == Qrange)
			{
			  range[i * 2] = parse_integer (nd.FirstChild);
			  range[i * 2 + 1] = parse_integer (nd.LastChild);
			}
		      else
			{
			  range[i * 2] = range[i * 2 + 1]
			    = parse_integer (nd.FirstChild);
			}
		    }
		}
	      vari = domain.DefvarInt (vname, val, desc, range);
	    }
	  else if (type == Qstring)
	    {
	      string[] range = null;
	      if (range_list != null)
		{
		  range = new string[nranges];
		  for (int i = 0; i < nranges; i++)
		    range[i] = range_list[i].FirstChild.InnerText;
		}
	      vari = domain.DefvarStr (vname, node.InnerText, desc, range);
	    }
	  else if (type == Qsymbol)
	    {
	      Symbol[] range = null;
	      if (range_list != null)
		{
		  range = new Symbol[nranges];
		  for (int i = 0; i < nranges; i++)
		    range[i] = range_list[i].FirstChild.InnerText;
		}
	      vari = domain.DefvarSym (vname, (Symbol) node.InnerText, desc,
				       range);
	    }
	  else
	    throw new Error (Error.TermTypeInvalid, "Unknown type: {0}", type);
	}
      else
	{
	  if (domain.variables.TryGetValue (vname, out vari))
	    vari = vari.Clone (domain);
	  else
	    vari = new Variable (domain, vname, Zero);
	  domain.variables[vname] = vari;
	}
      return vname;
    }

    private static Variable[] parse_args (Domain domain, XmlNode node,
					  out int min_args, out int max_args)
    {
      int nfixed = 0;
      int noptional = 0;
      int nrest = 0;
      Variable[] args;

      node = node.FirstChild;
      if (node != null && node.Name == Qargs)
	{
	  XmlNode n;
	  for (n = node.FirstChild; n != null; n = n.NextSibling)
	    {
	      if (n.Name == Qfixed)
		nfixed += n.ChildNodes.Count;
	      else if (n.Name == Qoptional)
		noptional += n.ChildNodes.Count;
	      else if (n.Name == Qrest)
		nrest++;
	      else
		throw new Error (Error.WrongType, n,
				 "Not fixed, optional, nor rest: {0}", n.Name);
	    }
	  min_args = nfixed;
	  max_args = nfixed + noptional + nrest;
	  args = new Variable[max_args];
	  n = node.FirstChild;
	  for (int i = 0; i < max_args; n = n.NextSibling)
	    {
	      for (XmlNode nn = n.FirstChild; nn != null; nn = nn.NextSibling)
		args[i++] = domain.Defvar ((Symbol) nn.Attributes[0].Value);
	    }
	  if (nrest == 1)
	    max_args = - max_args;
	}
      else
	{
	  min_args = max_args = 0;
	  args = new Variable[0];
	}
      return args;
    }

    public static Symbol parse_defun_head (Domain domain, XmlNode node)
    {
      Symbol name = node.Attributes[Qfname].Value;
      int min_args, max_args;
      Variable[] args;

      try {
	args = parse_args (domain, node, out min_args, out max_args);
	if (node.Name == Qdefun)
	  domain.Defun (name, min_args, max_args, args, null);
	else
	  domain.Defmacro (name, min_args, max_args, args, null);
      } catch (Error e) {
	if (e.ErrorNode == null)
	  e.ErrorNode = node;
	throw e;
      } catch (Exception e) {
	throw new Exception (node, e);
      }
      return name;
    }

    public static void parse_defun_body (Domain domain, XmlNode node)
    {
      Symbol name = node.Attributes[Qfname].Value;
      Function func = domain.GetFunc (name);

      for (node = node.FirstChild; node != null; node = node.NextSibling)
	if (node.Name != Qdescription
	    && node.Name != Qargs)
	  break;
      Term[] body = Parse (domain, node, null);
      if (node.Name == Qdefun)
	((Function.Lambda) func).SetBody (body);
      else
	((Function.Macro) func).SetBody (body);
    }

    public static Term Parse (Domain domain, XmlNode node)
    {
      Symbol name = node.Name;

      if (name == Qinteger)
	return new Term (parse_integer (node));
      if (name == Qsymbol)
	return new Term ((Symbol) node.InnerText);
      if (name == Qstring)
	return new Term (node.InnerText);
      if (name == Qlist)
	{
	  List<Term> list = new List<Term> ();
	  for (node = node.FirstChild; node != null;
	       node = node.NextSibling)
	    list.Add (Parse (domain, node));
	  return new Term (list);
	}
      if (name == Qdefun || name == Qdefmacro)
	{
	  name = parse_defun_head (domain, node);
	  parse_defun_body (domain, node);
	  return new Term (name);
	}
      if (name == Qdefvar)
	{
	  name = parse_defvar (domain, node);
	  return new Term (name);
	}
      MethodInfo minfo;
      if (domain.termtypes.TryGetValue (name, out minfo))
	{
	  object[] arg = new object[] { domain, node };
	  return new Term ((TermValue) minfo.Invoke (null, arg));
	}

      Funcall funcall = (Funcall) Funcall.Parser (domain, node);
      if (funcall.func is Function.Macro)
	return funcall.Eval (domain);
      return new Term (funcall);
    }

    private static bool default_stop (XmlNode n) { return n == null; }

    public static List<Term> Parse (Domain domain, XmlNode node,
				    Func<XmlNode, bool> stop)
    {
      if (stop == null)
	stop = default_stop;

      XmlNode n;
      for (n = node; ! stop (n); n = n.NextSibling)
	if (n.NodeType == XmlNodeType.Element
	    && (n.Name == Qdefun || n.Name == Qdefmacro))
	  parse_defun_head (domain, n);

      List<Term> terms = new List<Term> ();
      for (; node != n; node = node.NextSibling)
	if (node.NodeType == XmlNodeType.Element)
	  {
	    if (node.Name == Qdefun || node.Name == Qdefmacro)
	      parse_defun_body (domain, node);
	    else if (node.Name == Qdefvar)
	      parse_defvar (domain, node);
	    else
	      terms.Add (Parse (domain, node));
	  }
      return terms;
    }

    public static List<Term> Parse (Domain domain, string url)
    {
      XmlDocument doc = new XmlDocument (Symbol.NameTable);
      XmlNode node;

      using (XmlTextReader reader = new XmlTextReader (url, doc.NameTable))
	{
	  do {
	    reader.Read ();
	  } while (reader.NodeType != XmlNodeType.None
		   && (reader.NodeType != XmlNodeType.Element
		       || reader.Name != Qexpr));
	  if (reader.NodeType == XmlNodeType.None)
	    throw new Exception ("Node <expr> not found");
	  node = doc.ReadNode (reader);
	}
      return Parse (domain, node.FirstChild, null);
    }

    public static Term Eval (Domain domain, List<Term> terms)
    {
      Term result = Zero;

      foreach (Term term in terms)
	{
	  result = term.Eval (domain);
	  if (domain.Thrown)
	    return result;
	}
      return result;
    }
  }
}
